/* Copyright (C) |Meso|Star> 2015-2018 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d.h"
#include "s3d_backend.h"
#include "s3d_device_c.h"
#include "s3d_instance.h"
#include "s3d_shape_c.h"
#include "s3d_scene_c.h"

#include <rsys/float33.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
instance_release(ref_T* ref)
{
  struct instance* inst;
  struct s3d_scene* scn;
  ASSERT(ref);
  inst = CONTAINER_OF(ref, struct instance, ref);
  scn = inst->scene;
  MEM_RM(scn->dev->allocator, inst);
  S3D(scene_ref_put(scn));
}

/*******************************************************************************
 * Local functions
 ******************************************************************************/
res_T
instance_create
  (struct s3d_scene* scn,
   struct instance** out_inst)
{
  struct instance* inst = NULL;
  res_T res = RES_OK;
  ASSERT(scn && out_inst);

  inst = (struct instance*)
    MEM_CALLOC(scn->dev->allocator, 1, sizeof(struct instance));
  if(!inst) {
    res = RES_MEM_ERR;
    goto error;
  }
  f33_set_identity(inst->transform); /* rotation */
  f3_splat(inst->transform + 9, 0.f); /* Translation */
  ref_init(&inst->ref);
  S3D(scene_ref_get(scn));
  inst->scene = scn;
exit:
  *out_inst = inst;
  return res;
error:
  if(inst) {
    instance_ref_put(inst);
    inst = NULL;
  }
  goto exit;
}

void
instance_ref_get(struct instance* inst)
{
  ASSERT(inst);
  ref_get(&inst->ref);
}

void
instance_ref_put(struct instance* inst)
{
  ASSERT(inst);
  ref_put(&inst->ref, instance_release);
}

